// SimpleBinaryTree.cpp: implementation of the CSimpleBinaryTree class.
//
//////////////////////////////////////////////////////////////////////
/*
 *
 *
 *  Copyright (c) 2002 DigitalConvict <ax@digitalconvict.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * Contact info:
 * Site: http://www.digitalconvict.com
 * Email: ax@digitalconvict.com
 */

#include "stdafx.h"
#include "SimpleBinaryTree.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSimpleBinaryTree::CSimpleBinaryTree()
{
	m_nLastItemIndex=0;
	m_nTotalItems=0;
	m_nArraySize=0;
	m_nSizeOfItem=0;
	m_nDoBSThreshold=10;
	m_BTree=NULL;
	m_bAllowDuplicates=TRUE;
	m_bAutoResize=TRUE;
	m_bInitialised=FALSE;
	m_bFullySorted=FALSE;
}

/**********************************************************************
	Destructor:  Calls FreeMemory method;
**********************************************************************/
CSimpleBinaryTree::~CSimpleBinaryTree()
{
	FreeMemory();
}

/**********************************************************************
	Adds a new item to the BTree by allocating space for the new item
	and adding the pointer to the BTree array. Returns negative value
	on error.
**********************************************************************/
int CSimpleBinaryTree::AddItem(const void* pNewData, void* pStoredObject)
{
	ASSERT(m_bInitialised);

	int nResult;
	int nNewItemIndex;
	pStoredObject=NULL;

	if(m_bAutoResize 
		&& (int)((m_nTotalItems*100)/m_nArraySize)>=m_nGrowTrigger)
		ReSize(m_nArraySize+((m_nArraySize*m_nGrowByValue)/100));

	if(m_nLastItemIndex<(int)(m_nArraySize)){
		if(!m_bAllowDuplicates && FindItem(pNewData)>=0)
			return DUPLICATE_FOUND;

		if((m_BTree[m_nLastItemIndex]=malloc(m_nSizeOfItem))!=NULL){
			memcpy(m_BTree[m_nLastItemIndex], pNewData, m_nSizeOfItem);
			// increment total number of items in tree before any
			// sorting might be done - Thanks Steve V
			m_nTotalItems++;
			if(!m_bAllowDuplicates && m_nTotalItems>=m_nDoBSThreshold){
				SortItems();
				// use a full search if in Binary Search mode
				// as the index will most likely not be the last one
				// since we just sorted the tree - Thanks Steve V
				nNewItemIndex=FindItem(pNewData);
				pStoredObject=m_BTree[nNewItemIndex];
				nResult=nNewItemIndex;			
			} else {
				m_bFullySorted=FALSE;
				pStoredObject=m_BTree[m_nLastItemIndex];
				nResult=m_nLastItemIndex;
			}
			m_nLastItemIndex++;
		} else
			nResult=OUT_OF_MEMORY;
	} else
		nResult=TREE_IS_FULL;

	return nResult;
}

/**********************************************************************
	Decides which search method to use to find an item.  If the Array
	size is below the Binary Search Threshold then a linear search is
	performed instead of a Binary Search.
**********************************************************************/
int CSimpleBinaryTree::FindItem(const void* pvItemToFind, void* pvItemFound)
{
	ASSERT(m_bInitialised);

	int nResult;
	void **ptpItem=(void **)&pvItemToFind;
	pvItemFound=NULL;

	if(m_nTotalItems>m_nDoBSThreshold)
		nResult=BinarySearch(ptpItem);
	else
		nResult=LinearSearch(ptpItem);

	if(nResult!=ITEM_NOT_PRESENT)
		pvItemFound=m_BTree[nResult];

	return nResult;
}

/**********************************************************************
	Initialisation function MUST be called after object is created.
	Sets internal member variables and allocates memory for BTree array.
**********************************************************************/
BOOL CSimpleBinaryTree::Initialise(size_t nNumItems, size_t nItemSize,
									int (*compar)(const void *, 
										const void *),
									int nGrowTrigger,
									int nGrowByValue,
									int nShrinkTrigger,
									int nShrinkByValue)

{
	if(m_bInitialised)
		FreeMemory();

	if((m_BTree=new void*[nNumItems])==NULL)
		return FALSE;

	m_pfuncCompare=compar;
	m_nSizeOfItem=nItemSize;
	m_nArraySize=nNumItems;
	m_nGrowTrigger=nGrowTrigger;
	m_nGrowByValue=nGrowByValue;
	m_nShrinkTrigger=nShrinkTrigger;
	m_nShrinkByValue=nShrinkByValue;

	m_bInitialised=TRUE;

	return TRUE;
}

/**********************************************************************
	Simply sorts the elements in the array using C's quick sort function
**********************************************************************/
void CSimpleBinaryTree::SortItems()
{
	qsort(m_BTree, m_nTotalItems, sizeof(void *), m_pfuncCompare);
	m_bFullySorted=TRUE;
}

/**********************************************************************
	Shrinks or Grows the BTree according to the new Size indicated by
	nNewSize.  If nNewSize equals the current size or no more memory can
	be allocated the function returns FALSE otherwise returns TRUE.

	Elements of the array are copied to a new array and then deleted from
	the original array m_BTree;  m_BTree is then assigned the pointer to
	the	newly allocated array.
**********************************************************************/
BOOL CSimpleBinaryTree::ReSize(size_t nNewSize)
{
	ASSERT(m_bInitialised);

	BOOL bRetval=FALSE;
	void** pTempTree;
	pTempTree=new void*[nNewSize];

	if(pTempTree==NULL || nNewSize==m_nArraySize)
		return FALSE;

	int nLimit = (m_nTotalItems<(int)nNewSize ? m_nTotalItems : (int)nNewSize);

	memcpy(pTempTree, m_BTree, sizeof(void*)*nLimit);
	
	if((int)nNewSize<m_nTotalItems){
		for(int i=nLimit;i<m_nTotalItems;i++){
			delete m_BTree[i];
		}
	}

	delete[] m_BTree;
	m_BTree=pTempTree;
	m_nArraySize=nNewSize;

	if((int)nNewSize<m_nTotalItems)
		m_nTotalItems=nNewSize;

	return TRUE;
}

/**********************************************************************
	Returns a void* which points to the data item of the BTree indexed by
	nItem.  Returns NULL if nItem index is outside the valid range
**********************************************************************/
void* CSimpleBinaryTree::GetItemPtr(int nItem)
{
	ASSERT(m_bInitialised);

	if(nItem>=0 && nItem<m_nTotalItems)
		return m_BTree[nItem];
	else
		return NULL;
}

/**********************************************************************
	Frees up any heap memory allocated for the BTree array and items
	contained there within.  User should never have to call this unless
	they need memory in a hurry and are willing to destroy the BTree.

	Better to let the BTree go out of scope and have the default
	destructor call FreeMemory() automatically.
**********************************************************************/
void CSimpleBinaryTree::FreeMemory()
{
	ASSERT(m_bInitialised);

	for(int i=0;i<m_nTotalItems;i++){
		delete m_BTree[i];
	}
	if(m_BTree!=NULL){
		delete[] m_BTree;
		m_BTree=NULL;
		m_nLastItemIndex=0;
		m_nTotalItems=0;
	}
}

/**********************************************************************
	Removes an Item from the BTree. Item is deleted and all pointers 
	above this position in the array are shifted left 1 place. Item count
	member variable is updated accordingly. Returns TRUE if nItem is in
	the range of valid indexes.
**********************************************************************/
BOOL CSimpleBinaryTree::RemoveItem(int nItem)
{
	ASSERT(m_bInitialised);

	if(nItem>=0 && nItem<m_nTotalItems){
		delete m_BTree[nItem];
		for(int i=nItem+1;i<m_nTotalItems;i++){
			m_BTree[i-1]=m_BTree[i];
		}
		m_BTree[--m_nTotalItems]=NULL;
		m_nLastItemIndex--;

		if(m_bAutoResize 
			&& (int)((m_nTotalItems*100)/m_nArraySize)<=m_nShrinkTrigger)
			ReSize(m_nArraySize-((m_nArraySize*m_nShrinkByValue)/100));

		return TRUE;
	}

	return FALSE;
}

/**********************************************************************
	Performs a linear search on the elements of the m_BTree array.
	Returns the index of the element if the item is found or -1 otherwise
**********************************************************************/
int CSimpleBinaryTree::LinearSearch(const void* pvItemToFind)
{
	int nRetval=ITEM_NOT_PRESENT;

	for(int i=0;i<m_nTotalItems;i++){
		if(m_pfuncCompare(pvItemToFind, &m_BTree[i])==0){
			nRetval=i;
			break;
		}
	}

	return nRetval;
}

/**********************************************************************
	Performs Binary Search.  Returns index of Element if found in
	BTree otherwise returns ITEM_NOT_PRESENT if Element does not 
	exist in the tree
**********************************************************************/
int CSimpleBinaryTree::BinarySearch(const void *pvItemToFind)
{
	int nLeftPivot=0;
	int nRightPivot=m_nTotalItems-1;
	int nIndex;

	while(nLeftPivot<=nRightPivot){
		nIndex=nLeftPivot+(nRightPivot-nLeftPivot)/2;
		int nResult=m_pfuncCompare(pvItemToFind, &m_BTree[nIndex]);

		if(nResult==0)
			break;
		else if(nResult>0)
			nLeftPivot=nIndex+1;
		else
			nRightPivot=nIndex-1;
	}

	if(nLeftPivot>nRightPivot){
		if(!m_bFullySorted && m_bAllowDuplicates){
			// if we haven't found an item and the tree is unsorted
			// then sort it and search for the item again
			SortItems();
			nIndex=BinarySearch(pvItemToFind);
		} else 
			nIndex=ITEM_NOT_PRESENT;
	}

	return nIndex;
}

/**********************************************************************
	Sets the threshold value at which searching will be done using the
	binary search method instead of the linear search method.  The
	The search method is switched when the number of items in the tree
	exceeds the threshold value.  The default value is 10.
**********************************************************************/
void CSimpleBinaryTree::SetSearchMethodThreshold(int nThreshold)
{
	if(nThreshold>0){
		if(nThreshold<m_nDoBSThreshold && m_nTotalItems>nThreshold)
			SortItems();
		m_nDoBSThreshold=nThreshold;
	}
}

/**********************************************************************
	Gets the current threshold value.  See SetSearchMethodThreshold for
	more details.
**********************************************************************/
int CSimpleBinaryTree::GetSearchMethodThreshold()
{
	return m_nDoBSThreshold;
}

/**********************************************************************
	Gets the number of items currently stored in the tree.
**********************************************************************/
int CSimpleBinaryTree::GetTotalItems()
{
	return m_nTotalItems;
}

/**********************************************************************
	Gets the maximum number of items the tree can hold.
**********************************************************************/
int CSimpleBinaryTree::GetTreeSize()
{
	ASSERT(m_bInitialised);
	return (int)m_nArraySize;
}

/**********************************************************************
	Returns the current value at which a Growth Resize is triggered
**********************************************************************/
int CSimpleBinaryTree::GetGrowTriggerValue()
{
	ASSERT(m_bInitialised);
	return m_nGrowTrigger;
}

/**********************************************************************
	Sets the percentage value of the Grow Trigger.  When the Tree has
	allocated nPercentageUsed of its current space, the tree is expanded
	by automatically resizing if and only if AutoResizing is enabled.
	The default value for the GrowTrigger is 90%
**********************************************************************/
void CSimpleBinaryTree::SetGrowTriggerValue(int nPercentageUsed)
{
	ASSERT(m_bInitialised);
	if(nPercentageUsed>0)
		m_nGrowTrigger=nPercentageUsed%100;
}

/**********************************************************************
	Sets the percentage value of the Shrink Trigger.  When the Tree is
	using less than nPercentageUsed of its current space, the tree is
	truncated by automatically resizing if and only if AutoResizing is
	enabled. The default value for the ShrinkTrigger is 10%
**********************************************************************/
void CSimpleBinaryTree::SetShrinkTriggerValue(int nPercentageUsed)
{
	ASSERT(m_bInitialised);
	if(nPercentageUsed>0)
		m_nShrinkTrigger=nPercentageUsed%100;
}

/**********************************************************************
	Returns the current value at which a Shrink Resize is triggered
**********************************************************************/
int CSimpleBinaryTree::GetShrinkTriggerValue()
{
	ASSERT(m_bInitialised);
	return m_nShrinkTrigger;
}

/**********************************************************************
	Sets the percentage value by which the tree will grow if
	AutoResize is enabled.  When AutoResize is enabled and a Resize is
	triggered the tree will allocate nPercentIncrease items to the tree
**********************************************************************/
void CSimpleBinaryTree::SetGrowByPercentage(int nPercentIncrease)
{
	ASSERT(m_bInitialised);
	if(nPercentIncrease>0)
		m_nGrowByValue=nPercentIncrease%100;
}

/**********************************************************************
	Returns the current GrowBy Percentage
**********************************************************************/
int CSimpleBinaryTree::GetGrowByPercentage()
{
	ASSERT(m_bInitialised);
	return m_nGrowByValue;
}

/**********************************************************************
	Sets the percentage value by which the tree will shrink if
	AutoResize is enabled.  When AutoResize is enabled and a Resize is
	triggered the tree will deallocate nPercentDecrease empty items from
	the tree.
**********************************************************************/
void CSimpleBinaryTree::SetShrinkByPercentage(int nPercentDecrease)
{
	ASSERT(m_bInitialised);
	if(nPercentDecrease>0)
		m_nGrowByValue=nPercentDecrease%100;
}

/**********************************************************************
	Returns the current ShrinkBy Percentage
**********************************************************************/
int CSimpleBinaryTree::GetShrinkByPercentage()
{
	ASSERT(m_bInitialised);
	return m_nShrinkByValue;
}

/**********************************************************************
	Global generic 'compare' function provided for the user.  Takes
	two longs and compares them. Returns 0 if equal, negative int if a<b,
	postive	int if a>b.
**********************************************************************/
int GenericLongCompare(const void *a, const void *b)
{
	long *aa, *bb;

	aa=*(long**)a;
	bb=*(long**)b;

	return(*aa-*bb);
}

/**********************************************************************
	Provides example usage of the functions contained in the
	CSimpleBinarySearch Class.  This doesn't actually work unless you
	cut and paste the code in DoTest into your project
**********************************************************************/
//#define BINARY_TREE_EXAMPLE

#ifdef BINARY_TREE_EXAMPLE

// increase MAXELEMS to say 100 to use Binary Search Method
#define MAXELEMS 10
#define FINDLOOP 10

void DoTest(){
	// create btree object
	CSimpleBinaryTree btree;
	int index=0;
	int nThreshold=0;

	//Initialise btree and don't specify a compare function
	// cos we're going to use the default global one.
	btree.Initialise(MAXELEMS,sizeof(long));
	
	//Set Allow Duplicates flag to FALSE
	btree.m_bAllowDuplicates=FALSE;

	//Get Search Method Threshold  value
	nThreshold=btree.GetSearchMethodThreshold();

	// Turn AutoResize Off
//	btree.m_bAutoResize=FALSE;

	// Increment and set the search threshold value
	nThreshold+=2;
	btree.SetSearchMethodThreshold(nThreshold);

	// Add some random numbers to the btree,
	printf("\r\n---ADDING NEW ITEMS\r\n\r\n");

	for(int i=0;i<MAXELEMS;i++){
		int x=rand()%MAXELEMS;
		if((index=btree.AddItem(&x))>=0)
			printf("allocated integer with value %d into position %d\r\n", x, index);
		else {
			switch (index) {
			case OUT_OF_MEMORY:
				printf("NO MORE MEMORY AVAILABLE\r\n");
				break;
			case DUPLICATE_FOUND:
				printf("DUPLICATE: where item has value %d\r\n",x);
				break;
			case TREE_IS_FULL:
				printf("TREE FULL AT INDEX=%d\r\n",i);
				break;
			}
		}
	}

	//Find out if some values exist in the tree
	printf("\r\n---SEARCH FOR & RETRIEVE ITEMS\r\n\r\n");

	for(i=0;i<FINDLOOP;i++){
		int item=rand()%MAXELEMS;
		if((index=btree.FindItem(&item))>=0){
			printf("Found int with value %d at index %d\r\n", item, index);
		} else {
			printf("Integer with value %d not found\r\n", item);
		}

		if(index>=0){
			//Get a pointer to the data held in the btree
			printf("Getting pointer to data at index: %d\r\n", index);

			long *dataptr;
			if((dataptr=(long*)btree.GetItemPtr(index))!=NULL)
				printf("Value of data referenced by dataptr=%d\r\n\r\n", *dataptr);
			else
				printf("Got NULL Pointer. i.e. No Data Here!!\r\n\r\n");
		}
	}

	printf("\r\nResizing tree: doubling\r\n\r\n");
	// Resize the Btree
	btree.ReSize(MAXELEMS*2);

	printf("\r\n---FIND & REMOVE EXISTING ITEMS\r\n\r\n");
	//Find out if some values still exist in the tree
	for(i=0;i<FINDLOOP;i++){
		int item=rand()%MAXELEMS;
		if((index=btree.FindItem(&item))>=0)
			printf("Found integer with value %d at index %d\r\n", item, index);
		else
			printf("Integer with value %d not found\r\n", item);
		
		// Remove elements from the btree
		printf("Result of Removing item at index %d: %s\r\n", index,
			((btree.RemoveItem(index)) ? "OK" : "BAD INDEX"));
	}

	printf("\r\nResizing tree: reducing by half\r\n\r\n");
	// Resize the Btree
	btree.ReSize(MAXELEMS/2);
}

#endif

